/*
 *  Sniffer program to receive 802.15.4 packets
 *  can receive from Libelium Waspmote, intel iMote2 and Arduino
 *  can relay and serve as a relay node
 *
 *  Version:                0.38
 *  Design:                 C. Pham
 *  Implementation:         C. Pham
 */

/*
 *  Jun, 17th, 2013. v0.38
 *        adds 16-bit dest. address support, "D1234" set the dest. address to 1234 and set the XBEE short address to the last 4-digit of the XBee 64-bit MAC address, enabling
 *           (1) other motes working in 16-bit address mode to send packets to the XBee module (especially MicaZ and TelosB with CC2420 radio under TinyOS 
 *           (2) correct 16-bit source addr in the packets that will be forwarded by the node
 *        adds command "Y0#"/"Y1#" to disable/enable 16-bit node's short address
 *  Jun, 14th, 2013. v0.37
 *        adds command string prefix to '/@". All existing command should be prefixed such as: "/@Z10#"
 *  April, 25th, 2013. v0.36
 *        adds fast relay mode, enabled with FAST_RELAY    
 *  April, 6th, 2013. v0.35
 *        move preprocessor command into dynamic cmd for configuring the sniffer. IMPORTANT: 1 cmd at a time!:
 *          "CR#"/"CS# -> "CR#" reset timer0_millis to reset clock, "CS#" shows current millis() value
 *          "R1#"/"R0#" enable/disable relay mode;  
 *          "M1#"/"M0#" enable/disable miss pkt check;
 *          "PR1#"/"PR0#" enable/disable print received data;
 *          "PH1#"/"PH0#" enable/disable print received data in hexdump format;
 *          "D0013A2004076205E" to set destination address for relaying
 *          "BB#" or "BE#" or "BS20#" or "BC#"   
 *  March, 23rd, 2013. V0.34
 *        based on V0.34 for Libelium WaspMote
 */

/* TODO
 *       use of UART1 not working off-the-shelve as we need to have a HardwareSerial object for XBee 
 */

#include <XBee.h>
#include <SoftwareSerial.h>
#include <Streaming.h>

// BEGIN of compilation #define statements
//
// use advanced timing of the Libelium send API. CAUTION: need modified version of the API 
//#define SEND_API_TIMING
// fast relay mode, do not copy app data 
//#define FAST_RELAY
// print like hexdump any data that will be relayed
//#define PRINT_RLY_DATA
// get the number of ACK failures
//#define GET_ACK_FAILURES
// print detailed information
//#define TIMING_DISPLAY_VERBOSE
// indicates that the reception of packet uses the Libelium API 
// don't skip frame when multiple frames are in the rcv buffer, can lead to saturation of receive buffer
#define NO_SKIP_FRAME
// print the xbee raw data, prior to any processing such as escaped characters, for debug use mainly
//#define PRINT_XBEE_RAW_FRAME
// use the XBee Arduino lib for receiving packets
//#define USE_DEFAULT_XBEE_READ_LIB
// uses UART1 for receiving packet (and relaying in relay mode)
//#define USE_UART1
// use DigiMesh module for receiving data
//#define WITH_DIGIMESH
// uses an additional LCD display for more information
#define WITH_LCD
// uses 1 Led for packet relaying
//#define WITH_LED
// debug memory leaks
//#define DEBUG_MEMORY
//
// END of compilation #define statements

//The XBee RX (64 addr bit) API packet
#define XBEE_START_DELIMITER_BYTE        0
#define XBEE_LENGTH_MSB_BYTE             1
#define XBEE_LENGTH_LSB_BYTE             2
#define XBEE_API_IDENTIFIER              3
#define XBEE_SRC_ADDR_MSB                4
#define XBEE_SRC_ADDR_LSB                8
#define XBEE_RSSI                        12
#define XBEE_OPTION                      13
#define XBEE_DATA                        14

#define XBEE_DM_RESERVED                 12
#define XBEE_DM_RCV_OPTION               14
#define XBEE_DM_DATA                     15   

#define XBEE_CHECKSUM_SIZE               1
#define LIBELIUM_APP_HEADER_SIZE         9
#define IMOTE2_HEADER_SIZE               4

#define D802_RX_64                       0x80
#define D802_RX_16                       0x81
#define DM_RX                            0x90

#define MAX_DATA                         130
#define MAX_PARSE                        MAX_DATA

// relaying is done by broadcast by default
#define BROADCAST_MAC_ADDR       "000000000000FFFF"
#define SHORT_BROADCAST_MAC_ADDR "FFFF"
#define FINAL_DEST_MAC_ADDR       "0013A200408BC823"
#define SHORT_ADDR_FROM_64MAC     1
#define SHORT_ADDR_FFFF           2

//#define FINAL_DEST_MAC_ADDR "0013A200406FB3D3"
//#define FINAL_DEST_MAC_ADDR "0013A2004092204E"

#define SEND_LED                         DIGITAL1

// the maximum number of binary pkt that could be stored
#define MAX_NB_BIN_DATA_PKT              100

#ifdef USE_UART1
// not working off-the-shelve as we need to have a HardwareSerial object for XBee
SoftwareSerial mySerial(7, 8);
#endif

uint8_t* myRawData=NULL;
uint8_t myRawData_len;
uint8_t* myData=NULL;
uint8_t myData_len;

XBee xbee = XBee();

// useful AT commands
// write changes
uint8_t wrCmd[] = {'W','R'};
// get serial high
uint8_t shCmd[] = {'S','H'};
// get serial low
uint8_t slCmd[] = {'S','L'};
// get ack failures low
uint8_t eaCmd[] = {'E','A'};
// get cca failures low
uint8_t ecCmd[] = {'E','C'};
// get mac mode
uint8_t mmCmd[] = {'M','M'};
// set MY short address
uint8_t myAtCmd[] = {'M','Y'};

AtCommandRequest atRequest = AtCommandRequest();
AtCommandResponse atResponse = AtCommandResponse();

XBeeResponse response = XBeeResponse();

// create reusable response objects for responses we expect to handle 
#ifdef WITH_DIGIMESH
DMRxResponse rx64 = DMRxResponse();
DMTxStatusResponse txStatus = DMTxStatusResponse();
#else
Rx64Response rx64 = Rx64Response();
Rx16Response rx16 = Rx16Response();
TxStatusResponse txStatus = TxStatusResponse();
#endif

XBeeAddress64 destAddr64;
uint16_t destAddr16;

uint8_t option = 0;

String macAddr="";

#ifdef WITH_DIGIMESH
uint8_t XBEE_DATA_S=XBEE_DM_DATA;
#else
uint8_t XBEE_DATA_S=XBEE_DATA;
#endif

uint8_t XBEE_CURRENT_DATA_S;

// multi-purpose buffer for various information display
char banner[100];
// used mainly for the LCD display, typically remind the packet size and the generation frequency
char TGInfoStr[80];

uint16_t xbeeTotalPayload=0;
uint16_t appTotalPayload=0;

// will wrap at 255
uint8_t expectedSN=0;

int myMacMode=-1;

long nbReceivedPackets=0;
long nbLostPackets=0;
long startRelayTime, stopRelayTime, statusTime;
long startSendTime, stopSendTime, previousLastSendTime;
long lastReceivedTime,startTime,startFetchedTime,stopFetchedTime, startCopyTime, stopCopyTime, startBunchTime;
long lastSentTime=0;
unsigned long SnifferStartupTime;

// indicate that we received more than 1 frame in the serial buffer, need to continue with a fragmented frame
boolean nextFrame=false;
uint8_t skipFrame=0;

long nbRelayedPackets=0, nbRelayedPacketsError=0;

// broadcast by default
//char destMACAddr[17] = BROADCAST_MAC_ADDR; // need to reserve 1 byte more for the NULL-terminated character
char destMACAddr[17] = FINAL_DEST_MAC_ADDR; // need to reserve 1 byte more for the NULL-terminated character

// to manage binary pkt
boolean beginOfBinary=false;
boolean endOfBinary=false;
uint8_t* myBinaryData[MAX_NB_BIN_DATA_PKT];
uint8_t myBinaryData_len[MAX_NB_BIN_DATA_PKT];
int nbBinaryPkt=0;

boolean relay_mode=true;
boolean check_miss_pkt_mode=false;
boolean print_rcv_data=false;
boolean print_rcv_data_hex=false;
boolean have_short_address=false;

extern volatile unsigned long timer0_overflow_count;

#ifdef WITH_LCD
#include <Wire.h>
#include <LiquidCrystal_I2C.h>

LiquidCrystal_I2C lcd(0x20,16,2);  // set the LCD address to 0x20 for a 16 chars and 2 line display
#endif


void getTGInfo() {
  
        sprintf(TGInfoStr,"%s %s %s", relay_mode?"Relay":"Sniffer", check_miss_pkt_mode?"check":"nchck", have_short_address?(const char*)&macAddr[12]:SHORT_BROADCAST_MAC_ADDR);  
}

#ifdef WITH_LCD

void printLCD(int theLine, const char* theInfo, boolean clearFirst=true) {
        if (clearFirst)
              lcd.clear();
        lcd.setCursor(0, theLine);
        lcd.print(theInfo);
}

void printInfoLCD() {
        lcd.clear();
        getTGInfo();
        lcd.home();
        lcd.print(TGInfoStr);
        lcd.setCursor(0,1);
        printLCD(1, destMACAddr, false);
}

#endif

int freeMemory () {
  extern int __heap_start, *__brkval; 
  int v; 
  return (int) &v - (__brkval == 0 ? (int) &__heap_start : (int) __brkval); 
}

// used here to send AT command to the XBee to get information such as the MAC address
int sendAtCommand() {
#ifdef DEBUGPRINT
        Serial.println("Sending command to the Serial");
#endif
        // send the command
        xbee.send(atRequest);

        // wait up to 5 seconds for the status response
        if (xbee.readPacket(5000)) {
                // got a response!
                // should be an AT command response
                if (xbee.getResponse().getApiId() == AT_COMMAND_RESPONSE) {
                        xbee.getResponse().getAtCommandResponse(atResponse);
            
                        if (atResponse.isOk()) {
#ifdef DEBUGPRINT        
                                Serial.print("Command [");
                                Serial.print((char)atResponse.getCommand()[0]);
                                Serial.print((char)atResponse.getCommand()[1]);
                                Serial.println("] was successful!");
#endif
                                if (atResponse.getValueLength() > 0) {
#ifdef DEBUGPRINT          
                                        Serial.print("Command value length is ");
                                        Serial.println(atResponse.getValueLength(), DEC);
      
                                        Serial.print("Command value: 0x");
                                        
                                        for (int i = 0; i < atResponse.getValueLength(); i++) {
                                          Serial.print(atResponse.getValue()[i], HEX);
                                          Serial.print(" ");
                                        }
                              
                                        Serial.println("");
#endif          
                                }
                                return 1;
                        } 
                        else {
                                Serial.print("Command return error code: 0x");
                                Serial.println(atResponse.getStatus(), HEX);
                        }
                } 
                else {
                        Serial.print("Expected AT response but got 0x");
                        Serial.print(xbee.getResponse().getApiId(), HEX);
                }   
                
                
        } 
        else {
              // at command failed
              if (xbee.getResponse().isError()) {
                    Serial.print("Error reading packet.  Error code: 0x");  
                    Serial.println(xbee.getResponse().getErrorCode(), HEX);
              } 
              else {
                    Serial.print("No response from radio. ");  
              }
        }
        return 0;
}

#ifndef USE_DEFAULT_XBEE_READ_LIB

/********************************************************
 * int8_t getRawData(): read data from the serial buffer
 *
 *
 */ 
int8_t getRawData() {
        // get enough space to get a packet
        uint8_t* memory = (uint8_t*)calloc(MAX_DATA,sizeof(uint8_t));

        if (memory==NULL) {
                Serial.println("ERROR in getRawData:calloc:memory");
                Serial.flush();	// xbee.flush()	
                return 0;
        }
        
        uint16_t i=0;
        skipFrame=0;
        
        uint8_t b;
        long previous, previous2;
        long interval=50;
        long intervalMAX=40000;
        uint8_t maxFrame=30;
        // just let 5 millisecond max between 2 read, should be enough to detect the end of a frame
        interval = 5;
        
        boolean escapedChar=false;

// will leave next packets in the serial buffer for next getRawData() call
// allows for getting access to all data. 
// IMPORTANT: Will get buffer overwritting if traffic is too important!!
#ifdef NO_SKIP_FRAME       
 
        // did we leave a previous frame in the buffer from the last getRawData call?
        if (nextFrame) {
                // re-insert the frame delimiter
                memory[0]=0x7E;
                i=1;
                nextFrame=false;
                startFetchedTime=millis();
        }

        previous = millis();
        previous2 = previous; 
        
        // Read data from XBee meanwhile data is available
        // read until the serial buffer drains out, normally, for a reception, we should get the whole packet
        // at high rate, we can read more than 1 packet        
        while( ((millis()-previous) < interval) && ((millis()-previous2) < intervalMAX) && i < MAX_PARSE && !nextFrame) {	

                // xbee is linked on Serial so can use Serial here
                if (Serial.available()) { // xbee.available()

                        if (i==0)
                                startFetchedTime=millis();
                                
                        // xbee is linked on Serial so can use Serial here        
                        b = Serial.read(); // xbee.read();                        

#ifdef PRINT_XBEE_RAW_FRAME
                        Serial.print(b, HEX);
                        Serial.print(" ");
#endif                                
                        // start of a new frame?                        
                        // before the current frame is finishedn then truncated frame, we start over
                        if (b==0x7E && i>2 && i<memory[2]+3) {
                                nextFrame=true;
                                Serial.print("TRUNCATED");
                                free(memory);
                                return 0;
                        } 
                        
                        // we have to detect if 
                        // if it is really a new frame delimiter and we have actually read all the current frame, so leave it for next call to getRawData()    
                        if (b==0x7E && i>2 && i>memory[2]+3) {
                                nextFrame=true;
                        } 
                        else {
                                // ESCAPE?
                                if (b==0x7D) {
                                        // maybe we don't have a byte available yet
                                        if (Serial.available()) {
                                              // xbee is linked on Serial so can use Serial here
                                              b=Serial.read(); // xbee.read();         
#ifdef PRINT_XBEE_RAW_FRAME
                                              Serial.print(b, HEX);
                                              Serial.print(" ");
#endif   
                                              b = 0x20 ^ b;
                                        }
                                        else {
                                              escapedChar=true;
                                              continue;
                                        }      
                                }
                              
                                if (escapedChar==true)  {
                                    b = 0x20 ^ b;
                                    escapedChar=false;  
                                }
                                
                                memory[i]=b; 
                                i++;
                                previous=millis();
                        }
                }

                if ( millis()-previous < 0 ) previous=millis(); //avoid millis overflow problem
                if ( millis()-previous2 < 0 ) previous2=millis(); //avoid millis overflow problem
        }
        
#else  //end of #ifdef NO_SKIP_FRAME     
        
        // don't use this fast procedure when you are using relaying mode!
        
        // if more than 1 packet, just count the payload and the number of received frames because the frames are arriving too fast 
        // to leave them in the buffer. Give only access to the content of last packet
        previous = millis();
        previous2 = previous; 
        
        while( ((millis()-previous) < interval) && ((millis()-previous2) < intervalMAX) && i < MAX_PARSE) {	
                
                // did we leave a previous frame in the buffer?
                if (nextFrame) {
                        // re-insert the frame delimiter
                        memory[0]=0x7E;
                        nextFrame=false;
                        // reset
                        i=1;
                        startFetchedTime=millis();
                }

                // xbee is linked on Serial so can use Serial here
                if (Serial.available()) { // xbee.available()

                        if (i==0)
                                startFetchedTime=millis();

                        // xbee is linked on Serial so can use Serial here        
                        b = Serial.read(); // xbee.read();
 
                        // start of a new frame?
                        
                        // before the current frame is finishedn then truncated frame, we start over
                        if (b==0x7E && i>2 && i<memory[2]+3) {
                                nextFrame=true;
                                Serial.print("T");
                        } 
                        else                        
                                // if it is really a new frame delimiter and we have actually read all the current frame
                                if (b==0x7E && i>2 && i>memory[2]+3) {
                                        nextFrame=true;
                                        skipFrame++;
                                        Serial.print("S");
                                        
                                        int xbeedata=XBEE_DATA_S;
                                        
                                        if (memory[XBEE_API_IDENTIFIER]==D802_RX_16)
                                                xbeedata -= 6;
                                                
                                        // print the size only if the frame length is larger than expected, we have detected something wrong!
                                        if ( (i-XBEE_CHECKSUM_SIZE-xbeedata)!=memory[2]-11 ) 
                                                XBee.print(i-XBEE_CHECKSUM_SIZE-xbeedata);
                                        // update the amount of XBEE data received
                                        xbeeTotalPayload+=i-XBEE_CHECKSUM_SIZE-xbeedata;
                                        // here we assume that the TG runs without the Libelium API for sending
                                        // since this mode is only relevant for very high speed rate
                                        appTotalPayload+=i-XBEE_CHECKSUM_SIZE-xbeedata;
                                        
                                        if ((skipFrame % 40)==0) {
                                                double xbeeThroughput=xbeeTotalPayload*8/( (millis()-previous2)/1000.0 );
                                                double appThroughput=appTotalPayload*8/( (millis()-previous2)/1000.0 );                                      
                                                Serial.println("");
                                                Serial.print(i-XBEE_CHECKSUM_SIZE-xbeedata);
                                                Serial.print(" ");
                                                Serial.print(millis()-previous2);
                                                Serial.print(" X=");
                                                Serial.print(xbeeThroughput);
                                                Serial.print(" W=");
                                                Serial.print(appThroughput);      
                                                Serial.println("");  
                                                xbeeTotalPayload=0;
                                                appTotalPayload=0;  
                                                previous2 = millis();                                      
                                        }
                                } 
                                else {
                                        // ESCAPE?
                                        if (b==0x7D) {
                                                // maybe we don't have a byte available yet
                                                if (Serial.available()) {
                                                      // xbee is linked on Serial so can use Serial here
                                                      b=Serial.read(); // xbee.read();          
                                                      b = 0x20 ^ b;
                                                }
                                                else {
                                                      escapedChar=true;
                                                      continue;
                                                }      
                                        }
                                      
                                        if (escapedChar==true)  {
                                            b = 0x20 ^ b;
                                            escapedChar=false;  
                                        }
                                        
                                        memory[i]=b; 
                                        i++;
                                        previous=millis();
                                }
                }

                if ( millis()-previous < 0 ) previous=millis(); //avoid millis overflow problem
                if ( millis()-previous2 < 0 ) previous2=millis(); //avoid millis overflow problem
        }

#endif   
        // we got out of the while because of timing interval, not because there is no more frame
        if (nextFrame)
               skipFrame--; 
        if (skipFrame)
               Serial.println("");
               
        // we read something
        if (i) {
          
                XBEE_CURRENT_DATA_S=XBEE_DATA_S;
                                
                // we only keep RX event of 802.15.4 64-bit and DigiMesh
                if (memory[XBEE_API_IDENTIFIER] != D802_RX_64 && memory[XBEE_API_IDENTIFIER] != D802_RX_16 && memory[XBEE_API_IDENTIFIER] != DM_RX) {
                      free (memory);
                      myRawData=NULL;
                      return 0;
                }
                
                stopFetchedTime=millis();
#ifdef TIMING_DISPLAY_VERBOSE                
                Serial.print("tRCV(ms) ");
                Serial.print(millis());
                Serial.print(" tSL(ms) ");
                // the inter-arrival time
                Serial.print(startFetchedTime-lastReceivedTime);
                
                if (memory[XBEE_API_IDENTIFIER]==D802_RX_64) {
                        Serial.print(" RSSI(-dBm) ");
                        Serial.println((uint8_t)memory[XBEE_RSSI], DEC);
                }
                if (memory[XBEE_API_IDENTIFIER]==D802_RX_16) {
                        Serial.print(" RSSI(-dBm) ");
                        Serial.println((uint8_t)memory[XBEE_RSSI-6], DEC);
                        // with 16-bit addr receive, data start 6 bytes earlier
                        XBEE_CURRENT_DATA_S = XBEE_DATA_S-6;
                }    
                
                Serial.print("F(NO_L_API) ");
                // the total number of bytes fetched, the checksum is also fetched
                Serial.print(i);
                Serial.print("B in ");
                Serial.print(stopFetchedTime-startFetchedTime);
                Serial.print("ms. WAppPay ");
                Serial.print(i-XBEE_CHECKSUM_SIZE-XBEE_DATA_S);
                Serial.print("B. ");
#else
                // the absolute time
                Serial.print(startFetchedTime);
                Serial.print(" ");
                // the inter-arrival time
                Serial.print(startFetchedTime-lastReceivedTime);
                Serial.print(" FT(");
                Serial.print(i);
                Serial.print("):");
                Serial.print(stopFetchedTime-startFetchedTime);
#endif

                if (memory[XBEE_API_IDENTIFIER]==D802_RX_64) {
                  
                        if ((uint8_t)memory[XBEE_OPTION] & 0x02)
                                Serial.print(" -b. ");
                        else
                                Serial.print(" -u. ");
                }     
     
                if (memory[XBEE_API_IDENTIFIER]==D802_RX_16) {           
                  
                        if ((uint8_t)memory[XBEE_OPTION-6] & 0x02)
                                Serial.print(" -b. ");
                        else
                                Serial.print(" -u. "); 
                                
                        XBEE_CURRENT_DATA_S = XBEE_DATA_S-6;        
                }  
                
                
                if (memory[XBEE_API_IDENTIFIER]==DM_RX) {           
                  
                        if ((uint8_t)memory[XBEE_DM_RCV_OPTION]==0x02)
                                Serial.print(" -b. ");
                        else
                                Serial.print(" -u. ");
                }      

                // for next packet
                lastReceivedTime = startFetchedTime;
                            
                // set the myRawData pointer, one could extract additional information from the XBee header if required
                myRawData=memory;
                myRawData_len=i;
                
                int appDataStart;
                int dataStart;
                
                xbeeTotalPayload+=i-XBEE_CHECKSUM_SIZE-XBEE_CURRENT_DATA_S;
                // by default: a pure XBee (Arduino) without any additional header
                dataStart=XBEE_CURRENT_DATA_S;
#ifdef FAST_RELAY
                // set the length of the real data
                myData_len=i-1-appDataStart;
                // we set the pointer to the data
                myData = memory + dataStart;
                
                appTotalPayload+=myData_len;
                startCopyTime=millis();
#else
                appDataStart=dataStart;
                
                // test whether the packet comes from a WaspMote with the Libelium API
                if (memory[dataStart+2]=='#' && memory[dataStart+4]=='W' && memory[dataStart+5]=='A' && memory[dataStart+6]=='S' && memory[dataStart+7]=='P') {
                        // first fragment
                        appDataStart+=LIBELIUM_APP_HEADER_SIZE;
                }
                
                // test whether the packet comes from a WaspMote with the Libelium API, long packet fragmented in many small packets
                if (memory[dataStart+2]!='#' && memory[dataStart+3]=='W' && memory[dataStart+4]=='A' && memory[dataStart+5]=='S' && memory[dataStart+6]=='P') {
                        // not a first fragment
                        appDataStart+=LIBELIUM_APP_HEADER_SIZE-1;
                }                
                
                // probably from an iMote2 no TOS version, we assume that PANID=0x3332 which is put as payload by the iMote2 communication library
                if (memory[dataStart]==0x32 && memory[dataStart+1]==0x33) {
                        appDataStart+=IMOTE2_HEADER_SIZE;
                }

                // set the length of the real data
                myData_len=i-1-appDataStart;

                appTotalPayload+=myData_len;

                // allocate a new buffer to copy the real data, plus one char for the null-terminated character  
                myData = (uint8_t*) calloc(myData_len+1,sizeof(uint8_t));
                
                if (myData==NULL) {
                        Serial.println("ERROR in getRawData:calloc:myData");
                        Serial.print("try to allocate ");
                        Serial.println((int)myData_len+1);
                        
                        // xbee is linked on Serial so can use Serial here
                        Serial.flush(); // xbee.flush();		
                        if (memory) {
                                free(memory);
#ifdef DEBUG_MEMORY                                
                                Serial.println("free getRawData:myRawData");
#endif  
                                myRawData=NULL;
                        }
                        return 0;
                }

                startCopyTime=millis();
                for (int j=0; j<myData_len; j++) 
                        myData[j] = memory[appDataStart+j];  
#endif
                // set the null-terminated character so one could directly print the real data if it is a string
                myData[myData_len]='\0';
                stopCopyTime=millis();
                
                Serial.print("CT:");
                Serial.print(stopCopyTime-startCopyTime);
                // just to know, print the overhead of the getRawData() after stopFetchedTime
                Serial.print(" GRT:");
                Serial.println(millis()-stopFetchedTime);                
        }
        
        return i;
}
#endif

void setup() {
        SnifferStartupTime=millis();

        Serial.begin(38400);
        delay(500);
        Serial.print("Sniffer starts at ");   
        Serial.println(SnifferStartupTime);        
#ifdef WITH_LCD
        lcd.init();
        lcd.backlight();
        lcd.home();
        lcd.print("Init Sniffer");
        printLCD(1, "Init802", false);
        delay(500);
#endif            
        Serial.println("Init Sniffer. ");
        Serial.print(freeMemory());
        Serial.println(" bytes of free memory.");
        
#ifndef WITH_DIGIMESH
        if (relay_mode) {
              Serial.print("Relay to ");
              Serial.println(destMACAddr);
        }
        else
#endif
        {
              Serial.print("If requested will send img pkt to ");
              Serial.println(destMACAddr);
        }

#ifdef WITH_DIGIMESH
        Serial.print("Init XBee DigiMesh on ");
#else        
        Serial.print("Init XBee 802 on ");
#endif

#ifdef USE_UART1
        Serial.println("UART1");
#else
        Serial.println("UART0");
#endif   
        xbee.begin(38400);
        delay(500);
        
#ifdef USE_UART1
        xbee.setSerial(&mySerial);
#endif  
        // set command to SH to get high part of MAC address
        atRequest.setCommand(shCmd);  
        while (!sendAtCommand()) {Serial.println("retry command ATSH");};
 
        char adr_part[8];

        // get each value, convert it to string and append to the adr
        for (int i = 0; i < atResponse.getValueLength(); i++) {
              sprintf(adr_part, "%lX", (unsigned long)(atResponse.getValue()[i]));
    
              if (atResponse.getValue()[i]<=0xF) 
                macAddr += 0;
        
              macAddr += adr_part;
        } 
        
        // set command to SL to get low part of MAC address
        atRequest.setCommand(slCmd);  
        while (!sendAtCommand()) {Serial.println("retry command ATSL");};
  
        for (int i = 0; i < atResponse.getValueLength(); i++) {
              sprintf(adr_part, "%lX", (unsigned long)(atResponse.getValue()[i]));
    
              if (atResponse.getValue()[i]<=0xF) 
              macAddr += 0;
        
              macAddr += adr_part;    
        } 
        
#ifndef WITH_DIGIMESH        
        // set command to MM to get the XBee mac mode
        atRequest.setCommand(mmCmd);
        
        int counter = 4;
        while (!sendAtCommand() && counter > 0) {
              Serial.println("retry command ATMM");
              counter--;
        }
        
        if (counter > 0) {
                myMacMode=(int)atResponse.getValue()[0];
        }
        else
                Serial.print("ERROR AT+MM");

        sprintf(banner,"-mac:%s WAITING FOR PACKET FROM 802.15.4 devices. XBee mac mode %d", (const char*)&macAddr[0], myMacMode);
#else        
        sprintf(banner,"-mac:%s WAITING FOR PACKET FROM DigiMesh devices.", (const char*)&macAddr[0]);
#endif

        Serial.println(banner);

#ifdef WITH_LCD
        printLCD(1, (const char*)&macAddr[0], false);
        delay(2000);
        
        printInfoLCD();
#endif  

        lastReceivedTime = millis();     
        
        // Set params for relay
        if (strcmp(destMACAddr, BROADCAST_MAC_ADDR)==0) {
                // 64-bit addressing: This is the SH + SL address of remote XBee
                // unless you have MY on the receiving radio set to FFFF, this will be received as a RX16 packet
                destAddr64 = XBeeAddress64(0x00000000, 0x0000FFFF);        
        }
        else {
                // need to transform the MAC addr string into 2 uint32_t
                char myMacHigh[9], myMacLow[9];
                
                strncpy(myMacHigh, destMACAddr, 8);
                strncpy(myMacLow, destMACAddr+8, 8);
                myMacHigh[8]='\0';
                myMacLow[8]='\0';
                
                destAddr64 = XBeeAddress64((uint32_t)strtoul(myMacHigh, NULL, 16), (uint32_t)strtoul(myMacLow, NULL, 16));                    
        }        
}

void relayPkt(uint8_t SN, boolean binaryMode) {

        startRelayTime = millis();
        
        Serial.print("Relaying ");
        Serial.print(startRelayTime);

        if (strcmp(destMACAddr, BROADCAST_MAC_ADDR)==0 || strcmp(destMACAddr, SHORT_BROADCAST_MAC_ADDR)==0) {
                Serial.print(" -b");          
        }
        else {
                Serial.print(" -u");                          
        }          

        startSendTime=millis();
        previousLastSendTime=lastSentTime;
        lastSentTime=startSendTime;
        
#ifdef WITH_LED                
        digitalWrite(SEND_LED, HIGH);
#endif           
        
        uint8_t* origFrame;
        uint8_t data_len;
        
        if (beginOfBinary) {
                origFrame=myBinaryData[SN];
                data_len=myBinaryData_len[SN];
        }
        else {
                origFrame=myData;
                data_len=myData_len;
        }

#ifdef WITH_DIGIMESH
        // option is 0x00
        DMTxRequest tx = DMTxRequest(destAddr64, DM_BROADCAST_RADIUS_MAX_HOPS, 0x00, (uint8_t*)origFrame, data_len, SN);
#else
        statusTime=-1;
        
        // 16-bit destination address?        
        if (strlen(destMACAddr)==4) {
                // set the frame id to the pkt sequence number to keep track of ACK failures when mac mode is 0/2 with unicast address
                Tx16Request tx = Tx16Request(destAddr16, ACK_OPTION, (uint8_t*)origFrame, data_len, SN);
        
                // relay the frame
                xbee.send(tx);              
        }
        else {
                // set the frame id to the pkt sequence number to keep track of ACK failures when mac mode is 0/2 with unicast address
                Tx64Request tx = Tx64Request(destAddr64, ACK_OPTION, (uint8_t*)origFrame, data_len, SN);
        
                // relay the frame
                xbee.send(tx);          
        }
#endif
                   
        stopSendTime=millis();

#ifdef WITH_LED                
        digitalWrite(SEND_LED, LOW);
#endif 

#ifdef PRINT_RLY_DATA
        // binary pkt content is pure binary
        if (binaryMode) {
                // from stored binary pkt
                if (beginOfBinary) {
                        for (int k=0; k < myBinaryData_len[SN]; k++) {
                        
                                Serial.print((char)*(myBinaryData[SN]+k), HEX);
                                Serial.print(" ");
                        }
                }
                else {
                        for (int k=0; k < myData_len; k++) {
                        
                                Serial.print((char)myData[k], HEX);
                                Serial.print(" ");
                        }
                }
        }
        else        
                Serial.print((char*)myData);
                
        Serial.print(" ");
#endif

        // read incoming pkt, to get the transmit status, for 10ms max
        xbee.readPacket(10);

        if (xbee.getResponse().isAvailable()) {
#ifdef WITH_DIGIMESH
                // is it a response to the previously sent packet?            	
    	        if (xbee.getResponse().getApiId() == DM_TX_STATUS_RESPONSE) {
                      
                      statusTime=millis();
                      
    	              xbee.getResponse().getDMTxStatusResponse(txStatus);	
#else          
                // is it a response to the previously sent packet?            	
    	        if (xbee.getResponse().getApiId() == TX_STATUS_RESPONSE) {
                      
                      statusTime=millis();
                      
    	              xbee.getResponse().getTxStatusResponse(txStatus);	
#endif    
                
                      if (!txStatus.isSuccess()) {
                 
                            Serial.print("\nerror: 0x");                            
                            Serial.print(txStatus.getStatus(), HEX);
                            Serial.print(". Pkt ");
                            // get back the frame id
                            Serial.print(txStatus.getFrameData()[0]);
                            Serial.print(". ");
#ifdef GET_ACK_FAILURES                            
                            // set command to EA to get number of ACK failures
                            atRequest.setCommand(eaCmd);
        
                            int counter = 4;
                            while (!sendAtCommand() && counter > 0) {
                                  Serial.print("retry command ATEA");
                                  counter--;
                            }
                            
                            if (counter > 0) {
                                  Serial.print("0x");
                                  Serial.print(atResponse.getValue()[0], HEX);
                                  Serial.print(atResponse.getValue()[1], HEX);
                                  Serial.print(" ACK failures.");                              
                            }
                            else
                                  Serial.print("ERROR AT+EA."); 
                                  
                            // reset the ACK counter, we have to issue the equivalent of ATEA0
                            atRequest.setCommand(eaCmd);
                            atRequest.setCommandValue(0);
                            atRequest.setCommandValueLength(1);
                            
                            counter = 4;
                            while (!sendAtCommand() && counter > 0) {
                                  Serial.print("retry command ATEA0");
                                  counter--;
                            }           
                            
                            if (counter==0)
                                  Serial.print("ERROR AT+EA0.");
                            // clear the command value for next request
                            atRequest.clearCommandValue();                            
#endif                                  
                            nbRelayedPacketsError++;                           
                      }
                      else 
                            Serial.print("\nSuccess.");              
                }
                else
                      Serial.print("\nNot a TX Status.");
        }
        else 
              Serial.print("\nTimeout");    
              
        Serial.println("");
        Serial.print(startSendTime-previousLastSendTime);
        Serial.print(" ");                
        //Serial.print(". In ");
        Serial.println(stopSendTime-startSendTime);
        //Serial.print("ms. ");             
        
#ifdef SEND_API_TIMING
#ifdef TIMING_DISPLAY_VERBOSE
        sprintf(banner,"\nsend=%ld(%ld,%ld,%ld,%ld) parseMessage=%ld txStatusResponse=%ld\n",                   
#else
        sprintf(banner,"\n%ld %ld %ld %ld %ld %ld %ld\n",                     
#endif   
            xbee.stopSend-xbee.startSend,
            xbee.startWriteLoop-xbee.startSend,
            xbee.stopWriteLoop-xbee.startWriteLoop,
            xbee.startSendRadio-xbee.startSend,     
            xbee.stopSendRadio-xbee.startSendRadio,                                    
            xbee.stopParseMessage-xbee.startParseMessage,
            statusTime-stopSendTime);
        
        Serial.print(banner);    
#endif                                    
 
        nbRelayedPackets++;
        
        Serial.println("\n\n");
        stopRelayTime=millis();
}

void setMYAddr(int addrMode) {

        // set command to MY to set 16-bit node's short address
        atRequest.setCommand(myAtCmd); 
        uint16_t myShortAddr=0xFFFF;    
        
        Serial.print("Set node's short 16-bit short address to 0x");
        
        // we set the 16-bit addr by taking the last 4-digit of the 64-bit MAC address
        if (addrMode==SHORT_ADDR_FROM_64MAC) {

                Serial.println((const char*)&macAddr[12]);

                myShortAddr=(uint16_t)strtol((const char*)&macAddr[12], NULL, 16);     
                  
                uint8_t tmpValue;
                
                // need to swap byte order
                tmpValue = *((uint8_t*)&myShortAddr);
                *((uint8_t*)&myShortAddr) = *((uint8_t*)&myShortAddr+1);
                *((uint8_t*)&myShortAddr+1) = tmpValue;
                
                have_short_address = true;
        }
        else {   
                // otherwise, set to 0xFFFF to disable 16-bit address packet reception 
                Serial.print(SHORT_BROADCAST_MAC_ADDR);
                have_short_address = false;
        }
                
        atRequest.setCommandValue((uint8_t*)&myShortAddr);  
        
        atRequest.setCommandValueLength(sizeof(myShortAddr)); 

        while (!sendAtCommand()) {
                Serial.println("retry set command ATMY");
                delay(500);
        }
        
        atRequest.clearCommandValue();

        // set command to WR to write in EEPROM
        //atRequest.setCommand(wrCmd); 
        
        //while (!sendAtCommand()) {
        //        Serial.println("retry command ATWR");
        //        delay(500);
        //}
        
        // set command to read back MY 16-bit node's short address
        atRequest.setCommand(myAtCmd);    
        while (!sendAtCommand()) {
                Serial.println("retry read command ATMY");
                delay(500);
        }
        
        char my_part[4];
        String myAddr="";

        // get each value, convert it to string and append to the MY addr
        for (int i = 0; i < atResponse.getValueLength(); i++) {
              sprintf(my_part, "%lX", (unsigned long)(atResponse.getValue()[i]));
    
              if (atResponse.getValue()[i]<=0xF) 
                myAddr += 0;
        
              myAddr += my_part;
        }                                                         
        
        Serial.print("Node's short address is now: ");
        Serial.println((char*)&myAddr[0]);                        
}

void loop() {

#ifdef USE_DEFAULT_XBEE_READ_LIB // do not handle 16-bit address reception
       
        // read incoming pkt, mainly to get the transmit status
        xbee.readPacket();

        long startProcessing = millis();
        
        if (xbee.getResponse().isAvailable()) {
#ifdef WITH_DIGIMESH
                // is it a response to the previously sent packet?            	
    	        if (xbee.getResponse().getApiId() == DM_TX_STATUS_RESPONSE) {
                      
                      statusTime=millis();
                      
    	              xbee.getResponse().getDMTxStatusResponse(txStatus);	
#else          
                // is it a response to the previously sent packet?            	
    	        if (xbee.getResponse().getApiId() == TX_STATUS_RESPONSE) {
                      
                      statusTime=millis();
                      
    	              xbee.getResponse().getTxStatusResponse(txStatus);	
#endif                    
                      if (txStatus.isSuccess()) {  
#ifdef SEND_API_TIMING
#ifdef TIMING_DISPLAY_VERBOSE                         
                              sprintf(banner,"send=%ld(%ld,%ld,%ld,%ld) parseMessage=%ld txStatusResponse=%ld\n", 
#else
                              sprintf(banner,"\n%ld %ld %ld %ld %ld %ld %ld\n",
#endif                                
                                  xbee.stopSend-xbee.startSend,
                                  xbee.startWriteLoop-xbee.startSend,
                                  xbee.stopWriteLoop-xbee.startWriteLoop,
                                  xbee.startSendRadio-xbee.startSend,     
                                  xbee.stopSendRadio-xbee.startSendRadio,                                    
                                  xbee.stopParseMessage-xbee.startParseMessage,
                                  statusTime-stopSendTime);     
        
                              Serial.print(banner);
#endif                                               
                      }
                      else {
                            Serial.print("error: 0x");
                            Serial.print(txStatus.getStatus(), HEX);                
                            Serial.print(". Pkt ");
                            // get back the frame id
                            // normally, the one that was just sent 
                            Serial.print("NA");        
                            Serial.print(". ");
                                         
                            nbRelayedPacketsError++;
                      }
                }
                
#ifdef WITH_DIGIMESH
                if (xbee.getResponse().getApiId() == DM_RX_RESPONSE) {

                        xbee.getResponse().getDMRxResponse(rx64);
#else
                if (xbee.getResponse().getApiId() == RX_64_RESPONSE) {

                        xbee.getResponse().getRx64Response(rx64);
#endif                        
                        if (rx64.getErrorCode()==UNEXPECTED_START_BYTE) {
                                myRawData = NULL;
                                myData = NULL;
                                Serial.println("TRUNCATED");
                                return;
                        }
                        
                        option = rx64.getOption();
                        myRawData = rx64.getFrameData();
                        myRawData_len = rx64.getFrameDataLength();
                        myData = rx64.getData();
                        myData_len=rx64.getDataLength();
                        
                        xbeeTotalPayload+=myData_len;
                        appTotalPayload+=myData_len;
                        
                        stopFetchedTime=millis();
#ifdef TIMING_DISPLAY_VERBOSE                
                        Serial.print("tRCV(ms) ");
                        Serial.print(xbee.startFetchedTime);
                        Serial.print(" tSL(ms) ");
                        // the inter-arrival time
                        Serial.print(xbee.startFetchedTime-lastReceivedTime);

                        Serial.print(" RSSI(-dBm) ");
                        Serial.println(rx64.getRssi(), DEC);
                        
                        Serial.print("F(XBEE_API) ");
                        // the total number of bytes fetched, the checksum is also fetched
                        Serial.print(myRawData_len);
                        Serial.print("B in ");
                        Serial.print(xbee.stopFetchedTime-xbee.startFetchedTime);
                        Serial.print("ms. WAppPay ");
                        Serial.print(myData_len);
                        Serial.print("B. ");
#else
                        // the absolute time
                        Serial.print(xbee.startFetchedTime);
                        Serial.print(" ");
                        // the inter-arrival time
                        Serial.print(xbee.startFetchedTime-lastReceivedTime);
                        Serial.print(" FT(");
                        Serial.print(myRawData_len);
                        Serial.print("):");
                        Serial.print(xbee.stopFetchedTime-xbee.startFetchedTime);
#endif
                        if (rx64.isAddressBroadcast())
                                Serial.print(" -b. \n");
                        else
                                Serial.print(" -u. \n");   
                
                        lastReceivedTime=xbee.startFetchedTime;                       

#else // end of USE_DEFAULT_XBEE_READ_LIB

        // we use our own read procedure          
        // xbee is linked on Serial so can use Serial here
        if (Serial.available()) { // xbee.available()
                        
                long startProcessing = millis();
                
                // if getRawData() returns 0, it means that no byte could be read from the serial port 
                // or frame has been ignored
                // even if xbee.available() is true
                if (getRawData()) {                  
#endif 
                        long startDisplayRvcData = millis();
                        
                        if (nbReceivedPackets==0) {
                                startBunchTime=millis();
                                startTime=millis();
                        }        
                        
                        if (print_rcv_data)  {     
                                if (print_rcv_data_hex) {
                                        // print the HEX values like hexdump
                                        for (int j=0; j<myRawData_len; j++) {
                                                Serial.print(myRawData[j], HEX);
                                                Serial.print(" ");
                                               
                                                if (j==XBEE_DATA_S-1) {
                                                        Serial.print("||");
                                                        Serial.print(" ");
                                                }                                
                                        }
                
                                        Serial.println("");      
                                }
                                
                                if (!beginOfBinary)
                                        Serial.println((char*)(myData));      
                        }                  
                        else {
                                Serial.print("DLg:");
                                Serial.print((int)myData_len);
                                Serial.print("-");         
                        }
                        
                        Serial.print("DT:");
                        Serial.println(millis()-startDisplayRvcData);
                        
                        uint8_t seqN=-1; 
                        char* seqStr="           ";
                        
#ifdef FAST_RELAY
                        // with fast relay, we do not copy data into a separate memory space
                        boolean freeData=false;
#else
                        // by default, we have to free the received data
                        boolean freeData=true;
#endif
                        boolean already_relayed=false;

                        boolean isCmd=false;

                        // we use prefix '/@' to indicate a command string
                        if ((char)*myData=='/' && (char)*(myData+1)=='@') {
                          
                                // point to the first character after '$@'
                                char* myCmd=(char*)(myData+2);  
                          
                                // ONLY one command per input pkt!!!
                                switch ((char)*myCmd) {     
                                                                  
                                        // we assume that if the payload starts with S, it indicates a sequence number
                                        // for long messages that are fragmented by the Libelium API, the sequence number appears in the first fragment but not in the others
                                        // therefore, the next frgaments are not counted by nbReceivedPackets but their payload are counted for throughput computation
                                        case 'S':
                                                                                                                       
                                                // make sure that it is a sequence number and not a binary byte that has the 'S' value
                                                if ( (char)*(myCmd+1) >= '0' && (char)*(myCmd+1) <= '9') {
                                                
                                                        int i=1;
                                                        
                                                        // character '#' will indicate end of seq number
                                                        while ((char)*(myCmd+i)!='#' && i<strlen(seqStr)) {
                                                                *(seqStr+i-1)=*(myCmd+i);
                                                                i++;
                                                        }
                                                        *(seqStr+i-1)='\0';
                                                        seqN=atoi(seqStr);   
                                                        
                                                        if (check_miss_pkt_mode) {
                                                                if ((expectedSN+skipFrame)!=seqN) {
                                                                        nbLostPackets++;
                                                                        Serial.println("-MISS PKT-");
                                                                        Serial.println("");
                                                                }     
                                                        }                                              
                                        
#ifndef WITH_DIGIMESH        
                                                        if (relay_mode) {
                                                                // we only relay pkt with sequence number        
                                                                // skip the '#'
                                                                i++;
                                                                if ((char)*(myCmd+i)=='L') {
                                                                        // increase the level
                                                                        *(myCmd+i+1)=*(myCmd+i+1)+1;       
                                                                }                            
                                                                relayPkt(seqN, false);
                                                                Serial.print("SeqPkt RT:");
                                                                Serial.println(stopRelayTime-startRelayTime);
                                                                already_relayed=true;                                   
                                                        }
#endif 
                                                        // wait for next packet
                                                        if (seqN!=-1)
                                                                expectedSN=seqN+1;  
                                                }    
                                                break;
                                        
                                        // for clock synchronization
                                        // "CR#"/"CS#"
                                        case 'C':
                                                if ((char)*(myCmd+1)=='R') {
                                                        long old_millis=millis();
                                                        timer0_overflow_count=0;
                                                        delay(1000);
                                                        long new_millis=millis();
                                                        Serial.print("Old millis(): ");
                                                        Serial.print(old_millis);
                                                        Serial.print(" Clock reset, new millis() after 1000ms ");
                                                        Serial.println(new_millis);
                                                        lastReceivedTime=0;
                                                        lastSentTime=0;
                                                        startProcessing=0;
                                                        isCmd=true;                  
                                                }
                                                
                                                if ((char)*(myCmd+1)=='S') {
                                                        long current_millis=millis();
                                                        Serial.print("Clock show, current millis(): ");
                                                        Serial.println(current_millis);
                                                        isCmd=true;
                                                }
                                                
                                                break;
                                                
                                        // binary packets stream indicator?
                                        // "BB#" or "BE#" or "BS20#" or "BC#"  
                                        case 'B':
                                                
                                                if (!relay_mode) {
                                                        // start of binary stream? we may receive several for reliability concern
                                                        if ((char)*(myCmd+1)=='B') {
                                                                beginOfBinary=true;
                                                                Serial.println("Binary Begin");
                                                                isCmd=true;
                                                        }
                                                        
                                                        // end of binary stream? we may receive several for reliability concern
                                                        if ((char)*(myCmd+1)=='E') {
                                                                endOfBinary=true;
                                                                Serial.println("Binary End");
                                                                isCmd=true;
                                                        }    
                
                                                        if ((char)*(myCmd+1)=='C') {
                                                                Serial.println("Binary Cleared");       
                                                                
                                                                for (int i=nbBinaryPkt-1; i>=0; i--) 
                                                                        free(myBinaryData[i]);
                                                                nbBinaryPkt=0;
                                                                beginOfBinary=false;
                                                                endOfBinary=false;
                                                                isCmd=true;
                                                        }  
                                                        
                                                        // command to send the stored binary packet "BS20#": will use a delay of 20ms between each binary pkt
                                                        if ((char)*(myCmd+1)=='S') {
                                                                Serial.print("Binary Send with inter-pkt ");
                                                                
                                                                // character '#' will indicate end of inter-pkt time
                                                                int k=2;
                                                                
                                                                while ((char)*(myCmd+k)!='#' && ( (k-2)<strlen(seqStr) ) ) {
                                                                        *(seqStr+k-2)=*(myCmd+k);
                                                                        k++;
                                                                }
                                                                *(seqStr+k-1)='\0';
                                                                
                                                                int inter_binary_pkt=atoi(seqStr);
                                                                
                                                                Serial.print(inter_binary_pkt);
                                                                Serial.print(" to ");
                                                                Serial.println(destMACAddr);                                                  
                                                                
                                                                long now1, now2;
                                                                                            
                                                                for (int i=0; i<nbBinaryPkt; i++) {
                                                                        now1=millis();
                                                                        relayPkt(i, true);
                                                                        
                                                                        while ( (millis()-now1) < inter_binary_pkt)
                                                                                ;
                                                                }
                                                                
                                                                Serial.print("Sent ");
                                                                Serial.print(nbBinaryPkt);
                                                                Serial.println(" binary pkt");
                                                                isCmd=true;
                                                        }                                          
                                                }
                                                break;
                                                
                                          // command to set the destination, override the default destination "D0013A20040762053#" will send to 0013A20040762053  
                                          case 'D':  
                                                               
                                                if (  ((char)*(myCmd+1) >= '0' && (char)*(myCmd+1) <= '9') || ((char)*(myCmd+1) >= 'A' && (char)*(myCmd+1) <= 'F')) {
                                                        Serial.print("Set destination of binary pkt to ");    
        
                                                        // character '#' will indicate end address
                                                        int k=1;
                                                        
                                                        // we test for 16 digit maximum
                                                        while ( (char)*(myCmd+k)!='#' && ( (k-1)<strlen(BROADCAST_MAC_ADDR) ) ) {
                                                                *(destMACAddr+k-1)=*(myCmd+k);
                                                                k++;
                                                        }
#ifdef WITH_DIGIMESH
                                                        // DigiMesh does not support 16-bit address mode
                                                        if ((k-1)!=16) {
                                                                Serial.print("bad address, need 16 bytes in hex format. Restore default destination ");  
                                                                strcpy(destMACAddr, FINAL_DEST_MAC_ADDR);
                                                        }
#else
                                                        if ((k-1)!=16 && (k-1)!=4) {
                                                                Serial.print("bad address, need 16 or 4 bytes in hex format. Restore default destination ");  
                                                                strcpy(destMACAddr, FINAL_DEST_MAC_ADDR);
                                                        }
#endif                                                        
                                                        else        
                                                                *(destMACAddr+k-1)='\0';          
                                                            
                                                        Serial.println(destMACAddr);
                                                        
                                                        boolean setShortAddress=false;
                                                        
                                                        // 16-bit destination address
                                                        if (strlen(destMACAddr)==4) {
                                                                // to indicate that we need to set a short address for the node because 16-bit destination needs a node's short address as well
                                                                if (have_short_address==false)
                                                                        setShortAddress=true;
                                                                
                                                                if (strcmp(destMACAddr, SHORT_BROADCAST_MAC_ADDR)==0) {
                                                                        // 16-bit addressing: This is the SH + SL address of remote XBee
                                                                        // unless you have MY on the receiving radio set to FFFF, this will be received as a RX16 packet
                                                                        destAddr16 = 0xFFFF;    
                                                                } 
                                                                else {
                                                                        destAddr16=(uint16_t)strtol(destMACAddr, NULL, 16);
                                                                }                                 
                                                        }
                                                        else {
                                                                // we can have a 64-bit destination address and have a short address to enable 16-bit address addressing
                                                                // this is why we do not set haveShortAddress to false
                                                                if (strcmp(destMACAddr, BROADCAST_MAC_ADDR)==0) {
                                                                        // 64-bit addressing: This is the SH + SL address of remote XBee
                                                                        // unless you have MY on the receiving radio set to FFFF, this will be received as a RX16 packet
                                                                        destAddr64 = XBeeAddress64(0x00000000, 0x0000FFFF);   
                                                                }
                                                                else {
                                                                        // need to transform the MAC addr string into 2 uint32_t
                                                                        char myMacHigh[9], myMacLow[9];
                                                                        
                                                                        strncpy(myMacHigh, destMACAddr, 8);
                                                                        strncpy(myMacLow, destMACAddr+8, 8);
                                                                        myMacHigh[8]='\0';
                                                                        myMacLow[8]='\0';
                                                                        
                                                                        destAddr64 = XBeeAddress64((uint32_t)strtoul(myMacHigh, NULL, 16), (uint32_t)strtoul(myMacLow, NULL, 16));     
                                                                }           
                                                        }              
#ifndef WITH_DIGIMESH
                                                        if (setShortAddress) 
                                                                setMYAddr(SHORT_ADDR_FROM_64MAC);  
#endif                                                        
                                                        isCmd=true;
                                                }        
                                                break;

#ifndef WITH_DIGIMESH
                                          // command to set the node's short address to the last 4 digit of the 64-bit MAC address, "Y0#" or "Y1#" 
                                          case 'Y':  
                                                
                                                if ((char)*(myCmd+1)=='0')
                                                        setMYAddr(SHORT_ADDR_FFFF);
                                                else 
                                                        setMYAddr(SHORT_ADDR_FROM_64MAC);
                                                        
                                                isCmd=true;     
                                                                           
                                                break;
#endif                                                
                                          // print received data?
                                          // "PR1#" or "PR0#" or "PH1#" or "PH0"  
                                          case 'P':
                                                                                                
                                                // for received data?
                                                if ((char)*(myCmd+1)=='R') {
                                                        if ( (char)*(myCmd+2)=='0' || (char)*(myCmd+2)=='1') {
                                                  
                                                              print_rcv_data=((char)*(myCmd+2)=='0')?false:true;
                                                              if (print_rcv_data)
                                                                      Serial.println("Print rcv data");
                                                              else
                                                                      Serial.println("Do not print rcv data"); 
                                                              isCmd=true; 
                                                        }                                         
                                                }    
                                                
                                                // for received data?
                                                if ((char)*(myCmd+1)=='H') {
                                                        if ( (char)*(myCmd+2)=='0' || (char)*(myCmd+2)=='1') {
                                                  
                                                              print_rcv_data_hex=((char)*(myCmd+2)=='0')?false:true;
                                                              if (print_rcv_data)
                                                                      Serial.println("Print rcv data hexdump");
                                                              else
                                                                      Serial.println("Do not print rcv data hexdump");
                                                              isCmd=true;   
                                                        }                                         
                                                }                                            
                                                break;
                                                
                                         case 'R':
                                                if ( (char)*(myCmd+1)=='0' || (char)*(myCmd+1)=='1') {                                            
                                                      
                                                      // disable relay mode if cmdValue is 0, otherwise, we run in relay mode 
                                                      relay_mode=((char)*(myCmd+1)=='0')?false:true;
                                                      
                                                      if (relay_mode) {
                                                              Serial.print("Set to relay mode: ");
                                                              Serial.println(destMACAddr);
                                                      }
                                                      else {       
                                                              Serial.print("Set to sniffer mode");
                                                      }
                                                      Serial.println("");
                                                      isCmd=true;
                                                }
                                                break; 
        
                                         case 'M':
                                                if ( (char)*(myCmd+1)=='0' || (char)*(myCmd+1)=='1') {                                            
                                                      
                                                      // disable pkt sequence checking mode if cmdValue is 0, otherwise, we detect pkt miss 
                                                      check_miss_pkt_mode=((char)*(myCmd+1)=='0')?false:true;
                                                      
                                                      if (check_miss_pkt_mode) {
                                                              Serial.print("Set missed pkt check mode");
                                                      }
                                                      else {       
                                                              Serial.print("Don't check for missed pkt");
                                                      }
                                                      Serial.println("");
                                                      isCmd=true;
                                                }  
                                                break;
                                }   // end of switch                            
                        }        
                                                                                      
                        if (!isCmd) {
                                
                                // beginOfBinary is true so this is a binary packet that needs to be stored
                                if (beginOfBinary) {
#ifdef USE_DEFAULT_XBEE_READ_LIB                
                                        // allocate new memory space to stored this packet
                                        myData = (uint8_t*) calloc(myData_len+1,sizeof(uint8_t));
                                        
                                        if (myData==NULL) {
                                                Serial.println("ERROR in storePkt:calloc:myData");
                                                Serial.print("try to allocate ");
                                                Serial.println((int)myData_len+1);                                                        	
                                        }
                                        else {
                                                uint8_t* theData=rx64.getData();
                                                
                                                for (int j=0; j<myData_len; j++) 
                                                        myData[j] = theData[j];  
                                
                                                // set the null-terminated character so one could directly print the real data if it is a string
                                                myData[myData_len]='\0';
                                        }
                                        
#endif                          
                                        // without our own library we already have data in myData
                                        if (myData != NULL) {    
                                                myBinaryData[nbBinaryPkt]=myData;
                                                myBinaryData_len[nbBinaryPkt]=myData_len;                              
                                                Serial.print("Stored BinaryPkt ");
                                                Serial.println(nbBinaryPkt);
                                                nbBinaryPkt++;
                                                // indicate that myData should not be free-ed
                                                freeData=false;
                                        }        
                                        Serial.print(freeMemory());
                                        Serial.println(" B left.");     
                                }
                                else
                                        nbReceivedPackets++;                               
                                
                                // print statistics every 20 pkts
                                if ((nbReceivedPackets % 20)==0) {
                                        double xbeeThroughput=xbeeTotalPayload*8/( (millis()-startBunchTime)/1000.0 );
                                        double appThroughput=appTotalPayload*8/( (millis()-startBunchTime)/1000.0 );
                                        Serial.print("\n20pkt ");
                                        Serial.print(millis()-startBunchTime);
                                        Serial.print("ms. XBee=");
                                        Serial.print(xbeeThroughput);
                                        Serial.print("bps. App=");
                                        Serial.print(appThroughput);      
                                        Serial.println("bps.");           
                                        xbeeTotalPayload = 0;
                                        appTotalPayload = 0;
                                        
                                        Serial.print("Rcv ");
                                        Serial.print(nbReceivedPackets);
                                        Serial.print(". Lost ");
                                        Serial.println(nbLostPackets);
#ifndef WITH_DIGIMESH
                                        if (relay_mode) {
                                                Serial.print("Rly ");
                                                Serial.print(nbRelayedPackets);
                                                Serial.print("(");
                                                Serial.print(nbRelayedPacketsError);
                                                Serial.println(")");
                                        }
#endif                                                                          
                                        Serial.print("ST:");
                                        Serial.println(millis()-startTime); 
                                        Serial.println("");                               
        
                                        startBunchTime=millis();
                                }                                                               
#ifndef WITH_DIGIMESH
                                // only for what is considered binary pkt: those not beginning with a seq number (these are already handled)
                                // and those that are not seen as commands.
                                if (relay_mode && !already_relayed) {
                                      // relay the binary pkt
                                      relayPkt(nbBinaryPkt, true);
                                      Serial.print("BinaryPkt RT:");
                                      Serial.println(stopRelayTime-startRelayTime);                                        
                                      nbBinaryPkt++;
                                }
#endif           
                        }
                        else {
#ifdef WITH_LCD 
                                printInfoLCD();
#endif     
                        }
                                                
#ifndef USE_DEFAULT_XBEE_READ_LIB                                                                    
                        // free myRawData
                        free(myRawData);
                        myRawData=NULL;
#ifdef DEBUG_MEMORY                        
                        Serial.println("free myRawData");      
#endif         
                        // myData, note that we chose to make both memory buffers independent 
                        // if we have to store the image, then do not free the memory
                        // we will free later on when requested
                        if (freeData) {                        
                                free(myData);
#ifdef DEBUG_MEMORY                                
                                Serial.println("free myData");
#endif                                
                        }
#endif                      
                        myData=NULL; 
#ifdef DEBUG_MEMORY                        
                        Serial.print(freeMemory());
                        Serial.println(" B left.");
#endif                        
                        Serial.print("PT:");
#ifdef USE_DEFAULT_XBEE_READ_LIB
                        Serial.println(millis()-startProcessing+xbee.stopFetchedTime-xbee.startFetchedTime);
#else
                        Serial.println(millis()-startProcessing);
#endif                        
                        Serial.println(""); 
                }               
        }
}






